/*
    socklibpp library
    Copyright (C) 2005  Daniel K. O. <danielosmari at users.sf.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <string>
#include <istream>
#include <ostream>
#include <algorithm>
#include <functional>
#include <cstdlib>
#include <cstring>

#include "sockaddr.h"

#include "sockcommon-private.h"


using namespace std;

namespace socklibpp {



/**     Initializes the family to sock_addr_in::af_inet, the port to 0,
        and the address to sock_addr_in::addr_none.
*/
addr_in::addr_in()
throw() :
        addr(reinterpret_cast<struct sockaddr*>(&_M_addr), sizeof _M_addr)
{
        _M_addr.sin_family = af_inet;
        _M_addr.sin_port = 0;
        uint32_t temp = addr_none;
        memcpy(&_M_addr.sin_addr, &temp, 4);
}



/**     \param addr_    A default IPv4 address
        \param port_    A 16-bit IPv4 port
*/
addr_in::addr_in(address addr_, uint16_t port_)
throw() :
        addr(reinterpret_cast<struct sockaddr*>(&_M_addr), sizeof _M_addr)
{
        _M_addr.sin_family = af_inet;
        _M_addr.sin_port = htons(port_);
        uint32_t temp = htonl(addr_);
        memcpy(&_M_addr.sin_addr, &temp, 4);
}
        


/**     \param addr_    A 32-bit IPv4 address, in host-byte-order
        \param port_    A 16-bit IPv4 port, in host-byte-order
*/
addr_in::addr_in(uint32_t addr_, uint16_t port_)
throw() :
        addr(reinterpret_cast<struct sockaddr*>(&_M_addr), sizeof _M_addr)
{
        _M_addr.sin_family = af_inet;
        _M_addr.sin_port = htons(port_);
        addr_ = htonl(addr_);
        memcpy(&_M_addr.sin_addr, &addr_, 4);
}



/**
        \param str_     A string in the format "X.X.X.X", where "X" is an
                        integer in [0,255]
        \param port_    A 16-bit IPv4 port
*/
addr_in::addr_in(const string& str_, uint16_t port_)
throw (invalid_argument) :
        addr(reinterpret_cast<struct sockaddr*>(&_M_addr), sizeof _M_addr)
{
        _M_addr.sin_family = af_inet;

        if (!validate_addr(str_))
                throw invalid_argument("Invalid address for addr_in");

        uint32_t temp = ::inet_addr(str_.c_str());
        memcpy(&_M_addr.sin_addr, &temp, 4);
        _M_addr.sin_port = htons(port_);
}



/**
        \param str_     A string in the format "X.X.X.X:P", where "X" is an
                        integer in [0,255] and "P" is an integer in [0,65535]
*/
addr_in::addr_in(const string& str_)
throw (invalid_argument) :
        addr(reinterpret_cast<struct sockaddr*>(&_M_addr), sizeof _M_addr)
{
        _M_addr.sin_family = af_inet;

        if (!ip_port(str_))
                throw invalid_argument("Invalid address for addr_in");
}



void
addr_in::reset()
throw()
{
        _M_size = sizeof _M_addr;
}



bool
addr_in::validate_addr(const string& str)
throw()
{
        if (str.empty())
                return false;
                
        // avoid Winsock crash
        if (str.find_first_not_of("0123456789.") != string::npos)
                return false;
        
        return inet_addr(str.c_str()) != addr_none;
}



bool
addr_in::validate_addr_port(const string& str)
throw()
{
        if (str.empty())
                return false;
                
        if (str.find_first_not_of("0123456789.:") != string::npos)
                return false;

        string::size_type pos = str.find(':');
        
        if (pos == string::npos || pos == str.length()-1)
                return false;
        
        char *e;
        uint32_t port = strtoul(str.c_str()+pos+1, &e, 10);

        if (*e || port>65535)
                return false;
        
        return inet_addr(str.substr(0, pos).c_str()) != addr_none;
}



namespace {
bool
is_ascii_string(const string& addrport)
throw()
{
        if (find_if(addrport.begin(), addrport.end(), bind2nd(greater<unsigned char>(), 127))
        != addrport.end())
                return false;

        return true;
}
}



/**	This is thread safe; you can call the resolve() method from multiple
        threads without problems.
        
        \param str      A domain name to be resolved

        \return true if the name was resolved successfully

        \note   This method blocks until the resolving operation is done.
*/
bool
addr_in::resolve(const string &str)
throw()
{
        
        #ifdef HAVE_GETHOSTBYNAME_R
        // For *nices with GNU libc
        struct hostent hostbuf, *host = 0;
        char extrabuf[1024];
        int err;
        if (gethostbyname_r(str.c_str(), &hostbuf, extrabuf, sizeof extrabuf, &host, &err) || !host)
                return false;

        memcpy(&_M_addr.sin_addr, host->h_addr_list[0], min(int(host->h_length), size()));

        #else // Win32, Mac OS...


        # ifdef WIN32
        if (!is_ascii_string(str))      // Winsock crashes when resolving non-ascii strings
                return false;
        # endif
        struct hostent *host;
        host = gethostbyname( str.c_str() );
        if (!host)
                return false;

        memcpy(&_M_addr.sin_addr, host->h_addr_list[0], min(int(host->h_length), size()));

        #endif
        

        return true;
}



/**	\param addrport A string in the format "name:port", where port is in
                        [0,65535]

        \return true if the name was resolved successfully and the port is valid

        \note   This method blocks until the resolving operation is done.

        \sa resolve(), port()
*/
bool
addr_in::resolve_port(const string& addrport)
throw()
{
        string::size_type pos = addrport.find(':');
        
        if (pos == string::npos || pos == addrport.length()-1)
                return false;

        if (!port(addrport.substr(pos+1)))
                return false;

        if (!resolve(addrport.substr(0, pos)))
                return false;

        return true;
}



/**	\param a        A string with the domain name to be resolved
        \param p        A 16-bit IPv4 port

        \return true if the name was resolved successfully

        \note   This method blocks until the resolving operation is done.

        \sa resolve(), port()
*/
bool
addr_in::resolve_port(const string& a, uint16_t p)
throw()
{
        return port(p), resolve(a);
}



/**
        \param a        A string with the domain name to be resolved
        \param p        A string representing the port

        \return true if the name was resolved successfully and the port is valid

        \note   This method blocks until the resolving operation is done.

        \sa resolve(), port()
*/
bool
addr_in::resolve_port(const string& a, const string& p)
throw()
{
        return resolve(a) && port(p);
}



uint16_t
addr_in::port() const throw ()
{
	return ntohs(_M_addr.sin_port); 
}



void
addr_in::port(uint16_t _port) throw () 
{
	_M_addr.sin_port = htons(_port);
}



/**
        \param port_    A string representing a port number or a service name
        \param proto_   A string representing a protocol name where the given
                        service is valid
        \return true if the operatin was successful
*/
bool
addr_in::port(const string& port_, const string& proto_)
throw()
{
        // let's try converting to an integer
        char *e;
        long res = std::strtol(port_.c_str(), &e, 10);
        if (!*e && 0<=res && res<65536) {
                _M_addr.sin_port = htons(res);
                return true;
        }

        /* TODO (DanielKO#1#): Check if getservbyname crashes if using non-ascii
                       strings. */
        struct servent *serv=0;
        #ifdef HAVE_GETSERVBYNAME_R
        struct servent result;
        char extrabuf[1024];
        if (getservbyname_r(port_.c_str(), proto_.c_str(), &result, extrabuf, sizeof extrabuf, &serv))
                serv = 0;
        else
                serv = &result;
        #else
        serv = getservbyname(port_.c_str(), proto_.c_str());
        #endif

        if (!serv)
        	return false;

        _M_addr.sin_port = serv->s_port;
        return true;
}


/**
        \param p        A string representing a port number or a service name
        \result true if the operation ws successful
*/
bool
addr_in::port(const string& p)
throw()
{
        if (p.empty())
                return false;

        // let's try converting to an integer
        char *e;
        long res = std::strtol(p.c_str(), &e, 10);
        if (!*e && 0<=res && res<65536) {
                _M_addr.sin_port = htons(res);
                return true;
        }

        /* TODO (DanielKO#1#): Check if getservbyname crashes if using non-ascii
                       strings. */

        struct servent *serv=0;
        #ifdef HAVE_GETSERVBYNAME_R
        struct servent result;
        char extrabuf[1024];
        if (getservbyname_r(p.c_str(), 0, &result, extrabuf, sizeof extrabuf, &serv))
                serv = 0;
        else
                serv = &result;
        #else
        serv = getservbyname(p.c_str(), 0);
        #endif

        if (!serv)
		return false;

        _M_addr.sin_port = serv->s_port;
        return true;
}



uint32_t
addr_in::ip() const throw()
{
	uint32_t temp;
        memcpy(&temp, &_M_addr.sin_addr, 4);
        return ntohl(temp);
}



/**
	\param host	A 32-bit unsigned integer representing the ip, in
			host-byte order.
*/
void
addr_in::ip(uint32_t host) throw()
{
        host = htonl(host);
        memcpy(&_M_addr.sin_addr, &host, 4);
}



/**
        \param addr_    A string in the format "X.X.X.X" where "X" is a number
                        in [0,255]

        \return true if the string is valid
*/
bool
addr_in::ip(const string& addr_)
throw()
{
        if (addr_.empty()) {
                const uint32_t temp = addr_none;
                memcpy(&_M_addr.sin_addr, &temp, 4);
                return true;
        }

        if (!validate_addr(addr_))
                return false;
                
        uint32_t temp = inet_addr(addr_.c_str());
        memcpy(&_M_addr.sin_addr, &temp, 4);
                
        return temp != addr_none;
}


/**
        \param addr_    A string in the format "X.X.X.X" where "X" is a number
                        in [0,255]
        \param port_     A 16-bit port
        \return true if the string is valid
        \sa ip(), port()
*/
bool
addr_in::ip_port(const string& addr_, uint16_t port_)
throw()
{
        return port(port_), ip(addr_);
}


/**
        \param addr_    A string in the format "X.X.X.X" where "X" is a number
                        in [0,255]
        \param port_     A string representing the port
        \return true if the address and the port are valid
        \sa ip(), port()
*/
bool
addr_in::ip_port(const string& addr_, const string& port_)
throw()
{
        return port(port_) && ip(addr_);
}


/**	\param addrp    A string in the format "X.X.X.X:P" where "X" is a number
                        in [0,255] and "P" is a number in [0,65535]

        \return true if the string is valid

        \sa ip(), port()
*/
bool
addr_in::ip_port(const string& addrp)
throw()
{
        if (!validate_addr_port(addrp))
                return false;
                
        string::size_type pos = addrp.find(':');
        if (pos == string::npos || pos==addrp.length()-1)
                return false;

        return ip(addrp.substr(0, pos)) && port(addrp.substr(pos+1));
}



/**	Returns a string in the format X.Y.Z.W. Notice that it doesn't include
	the port.
*/
string
addr_in::ip_str() const
{
	return string(inet_ntoa(_M_addr.sin_addr));
}



/**     Converts a addr_in object in a string representation, in the form:
        \a X.Y.Z.W:P
*/
ostream& operator<<(ostream& os, const addr_in& si)
{
        return os << si.ip_str() << ':' << ntohs(si._M_addr.sin_port);
}


/**     Reads a addr_in object from a string representation, in the form:
        \a X.Y.Z.W:P
*/
std::istream& operator>>(istream& is, addr_in& sai)
{
        if (!(is >> ws))
                return is;
        if (!is.good())
                return is;

        if (is.peek()==':') {
                is.clear(ios_base::badbit);
                return is;
        }

        string str;
        char ch;

        while (is.get(ch) && ch!=':')
                str += ch;

        if (!is)
                return is;

        uint16_t port;
        if (! (is >> port) )
                return is;

        if (! sai.ip_port(str, port) )
                is.clear(ios_base::badbit);
        
        return is;
}

}
